#!/usr/bin/env python3
#
# author: Claudio Driussi
#
"""
Produce una fattura pa in formato xml secondo le direttive a partire da un file 
json prodotto da Zefiro.

Si possono passare come parametri il file json da leggere ed il percorso di 
salvataggio del file xml.

La procedura e' in grado di leggere il file AGC.ini nella cartella corrente o 
nella cartella dello script per leggere parametri di configurazione, 
attualmente legge la possibilita' di togliere il registro dal nome del file. 
(VARIE,PYFPAREG)

"""
import xml.dom.minidom
import configparser as cp
import sys
import os
import json
import shutil
import base64

class FatPA(object):

    def __init__(self, file_in, ini):
        self.ini = ini
        if not self.ini.has_section('FatturaPA'):
            self.ini.add_section('FatturaPA')
        self.ini.set('DEFAULT','Art73','SI')
        self.ini.set('DEFAULT','TerzoIntermediario','')
        self.ini.set('DEFAULT','AddDescriz','')
        self.ini.set('DEFAULT','Causale','')
        self.ini.set('DEFAULT','Vers16','0')
        self.ini.set('DEFAULT','AddRdoc','0')
        self.ini.set('DEFAULT','AlIVA','0')
        self.ini.set('DEFAULT','DecPrunit','7')

        with open(file_in) as json_file:
            s = json_file.read()
            self.json = json.loads(s)

        self.sempl = self.json['fat_semplificata']
        self.autofat = self.json['t_doc'] in "TD16 TD17 TD18 TD19 TD20 TD21 TD22 TD23"

        self.doc = xml.dom.minidom.Document()
        self.escluso = self.json['cli_dest'] == 'NO'

        self.fuori_campo = 'N2'
        if self.json['d_doc'] > '2020-12-31' or self.ini.get('FatturaPA', 'Vers16') == '1':
            self.fuori_campo = 'N2.2'

        if self.sempl:
            self.semplificata()
        else:
            self.fatpa()

    def fatpa(self):
        # riconosco le PA perche' hanno il codice di interscambio lungo 6 caratteri
        versione = 'FPA12' if len(self.json['cli_dest']) == 6 else 'FPR12'

        # namespace fattura
        el = self.doc.createElementNS('http://ivaservizi.agenziaentrate.gov.it/docs/xsd/fatture/v1.2', 'p:FatturaElettronica')
        el.setAttribute("versione", versione)
        el.setAttribute("xmlns:ds", "http://www.w3.org/2000/09/xmldsig#")
        el.setAttribute("xmlns:p", "http://ivaservizi.agenziaentrate.gov.it/docs/xsd/fatture/v1.2")
        el.setAttribute("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance")
        el.setAttribute("xsi:schemaLocation", "http://ivaservizi.agenziaentrate.gov.it/docs/xsd/fatture/v1.2 fatturaordinaria_v1.2.xsd ")
        self.doc.appendChild(el)
        self.hdr = el.appendChild(self.doc.createElement("FatturaElettronicaHeader"))
        self.fat = el.appendChild(self.doc.createElement("FatturaElettronicaBody"))

        t = self.trasmissione(versione)

        if self.json['Tel'] or self.json['Email']:
            i = self.addElement('ContattiTrasmittente', t)
            if self.json['Tel']:
                x = self.addElement('Telefono', i, self.json['Tel'])
            if self.json['Email']:
                x = self.addElement('Email', i, self.json['Email'])

        # se il cessionario possiede una partita iva italiana e non e' iscritto
        # a SdI invio la PEC
        if self.json['cli_mail']:
            if self.json['cli_dest'] == '0000000' and self.json['cli_piva'] and not self.escluso:
                i = self.addElement('PECDestinatario', t, self.json['cli_mail'])
                
        # header: cedente prestatore
        t = self.addElement('CedentePrestatore', self.hdr)
        i0 = self.addElement('DatiAnagrafici', t)
        if self.json['frn_piva']:
            i1 = self.addElement('IdFiscaleIVA', i0)
            i2 = self.addElement('IdPaese', i1, self.json['frn_ivacee'])
            i2 = self.addElement('IdCodice', i1, self.json['frn_piva'])
            #  se c'e' un codice fiscale valido diverso dalla partita iva lo includo sempre
            if self.json['frn_codfis'] and self.json['frn_codfis'].strip() != self.json['frn_piva']:
                i1 = self.addElement('CodiceFiscale', i0, self.json['frn_codfis'])
        else:
            i1 = self.addElement('CodiceFiscale', i0, self.json['frn_codfis'])

        i1 = self.addElement('Anagrafica', i0)
        if self.json['TipoPers'] == 'S' or self.json['RagSoc'].strip():
            i2 = self.addElement('Denominazione', i1, self.json['RagSoc'])
        else:
            i2 = self.addElement('Nome', i1, self.json['Nome'])
            i2 = self.addElement('Cognome', i1, self.json['Cogn'])
        i1 = self.addElement('RegimeFiscale', i0, self.json['RegFis'])
        i0 = self.addElement('Sede', t)
        i1 = self.addElement('Indirizzo', i0, self.json['via'])
        i1 = self.addElement('CAP', i0, self.json['cap'])
        i1 = self.addElement('Comune', i0, self.json['citta'])
        i1 = self.addElement('Provincia', i0, self.json['prov'])
        i1 = self.addElement('Nazione', i0, self.json['frn_ivacee'])  
        if self.json['REA']: 
            i0 = self.addElement('IscrizioneREA', t)
            i1 = self.addElement('Ufficio', i0, self.json['REA'].split('-')[0])
            i1 = self.addElement('NumeroREA', i0, self.json['REA'].split('-')[1])
            if self.json['Capitale']: 
                i1 = self.addElement('CapitaleSociale', i0, self.json['Capitale'], '%.2f')
            i1 = self.addElement('SocioUnico', i0, 'SU' if self.json['SocioUnico'] else 'SM')
            i1 = self.addElement('StatoLiquidazione', i0, 'LN')            
        if self.json['Tel'] or self.json['Fax'] or self.json['Email']:
            if not self.autofat:
                i0 = self.addElement('Contatti', t)
                if self.json['Tel']:
                    i1 = self.addElement('Telefono', i0, self.json['Tel'])
                if self.json['Fax']:
                    i1 = self.addElement('Fax', i0, self.json['Fax'])
                if self.json['Email']:
                    i1 = self.addElement('Email', i0, self.json['Email'])

        # header: cessionario committente
        t = self.addElement('CessionarioCommittente', self.hdr)
        i0 = self.addElement('DatiAnagrafici', t)
        if self.json['cli_piva']:
            i1 = self.addElement('IdFiscaleIVA', i0)
            i2 = self.addElement('IdPaese', i1, self.json['cli_ivacee'])
            i2 = self.addElement('IdCodice', i1, self.json['cli_piva'])
            #  se c'e' un codice ficale valido diverso dalla partita iva lo includo sempre
            if self.json['cli_codfis'] and self.json['cli_codfis'].strip() != self.json['cli_piva']:
                i1 = self.addElement('CodiceFiscale', i0, self.json['cli_codfis'])
        else:
            i1 = self.addElement('CodiceFiscale', i0, self.json['cli_codfis'])
        i1 = self.addElement('Anagrafica', i0)
        i2 = self.addElement('Denominazione', i1, self.json['cli_ragsoc'])
        i0 = self.addElement('Sede', t)
        i1 = self.addElement('Indirizzo', i0, self.json['cli_via'])
        cap = self.json['cli_cap'] if self.json['cli_cap'] else '00000'
        i1 = self.addElement('CAP', i0, cap)
        i1 = self.addElement('Comune', i0, self.json['cli_citta'])
        i1 = self.addElement('Provincia', i0, self.json['cli_prov'])
        i1 = self.addElement('Nazione', i0, self.json['cli_ivacee'])

        # header: terzo intermediario
        # nel file ini e' presente una stringa che viene trasformata in un 
        # dizionario
        # es. CF:DRSCLD56M29F750D;RAGSOC:ROSSI SRL;EORI:01231280932;SOGGETTO:TZ
        # da :
        tz = self.ini.get('FatturaPA', 'TerzoIntermediario')
        if tz:
            tz = tz.split(';')
            dterzo = {}
            for i in tz:
                x = i.strip().split(':')
                dterzo[x[0]] = x[1]
            t = self.addElement('TerzoIntermediarioOSoggettoEmittente', self.hdr)
            i0 = self.addElement('DatiAnagrafici', t)
            i1 = self.addElement('CodiceFiscale', i0, dterzo['CF'])
            i1 = self.addElement('Anagrafica', i0)
            i2 = self.addElement('Denominazione', i1, dterzo['RAGSOC'])
            i2 = self.addElement('CodEORI', i1, dterzo['EORI'])
            t = self.addElement('SoggettoEmittente', self.hdr, dterzo['SOGGETTO'])

        # body: Dati generali documento
        gen = self.addElement('DatiGenerali', self.fat)
        i0 = self.addElement('DatiGeneraliDocumento', gen)
        i1 = self.addElement('TipoDocumento', i0, self.json['t_doc'])
        i1 = self.addElement('Divisa', i0, 'EUR')
        i1 = self.addElement('Data', i0, self.json['d_doc'])
        if self.json['r_doc'] == 'A' and self.ini.get('FatturaPA', 'AddRdoc') == '0':
            i1 = self.addElement('Numero', i0, self.json['n_doc'], '%d')
        else:
            i1 = self.addElement('Numero', i0, "%d/%s" % (self.json['n_doc'],self.json['r_doc']))
        if 'Ritenuta' in self.json:
            i1 = self.addElement('DatiRitenuta', i0)
            i2 = self.addElement('TipoRitenuta', i1, self.json['RitTipoPers'])
            i2 = self.addElement('ImportoRitenuta', i1, self.json['Ritenuta'], '%.2f')
            i2 = self.addElement('AliquotaRitenuta', i1, self.json['RitAliquota'], '%.2f')
            i2 = self.addElement('CausalePagamento', i1, self.json['RitCausale'])
        if 'Enasarco' in self.json:
            i1 = self.addElement('DatiCassaPrevidenziale', i0)
            i2 = self.addElement('TipoCassa', i1, 'TC07')
            i2 = self.addElement('AlCassa', i1, self.json['EnasarcoAlCassa'], '%.2f')
            i2 = self.addElement('ImportoContributoCassa', i1, self.json['Enasarco'], '%.2f')
            i2 = self.addElement('AliquotaIVA', i1, 0, '%.2f')
            i2 = self.addElement('Natura', i1, self.fuori_campo)
        if 'bollo' in self.json:
            i1 = self.addElement('DatiBollo', i0)
            i2 = self.addElement('BolloVirtuale', i1, 'SI')
            i2 = self.addElement('ImportoBollo', i1, self.json['bollo'], '%.2f')
        if 'tipo' in self.json:
            i1 = self.addElement('ScontoMaggiorazione', i0)
            i2 = self.addElement('Tipo', i1, self.json['tipo'])
            if self.json['sconto']:
                i2 = self.addElement('Percentuale', i1, self.json['sconto'], '%.2f')
            i2 = self.addElement('Importo', i1, self.json['scimpo'], '%.2f')
        i1 = self.addElement('ImportoTotaleDocumento', i0, self.json['totale'], '%.2f')
        if 'abbuono' in self.json:
            i1 = self.addElement('Arrotondamento', i0, self.json['abbuono'], '%.2f')
        if self.ini.get('FatturaPA', 'Causale'):
            i1 = self.addElement('Causale', i0, self.ini.get('FatturaPA', 'Causale'))
        if self.ini.get('FatturaPA', 'Art73') == 'SI':
            i1 = self.addElement('Art73', i0, 'SI')

        if 'DatiFattureCollegate' in self.json:
            i0 = self.addElement('DatiFattureCollegate', gen)
            i1 = self.addElement('IdDocumento', i0, self.json['DatiFattureCollegate'])
            if 'DatiFattureCollegate_Data' in self.json:
                i1 = self.addElement('Data', i0, self.json['DatiFattureCollegate_Data'])

        # dal 2023-04-06 le fatture collegate alle note d'accredito non funzionano piu'
        # sembra che se c'e' "DatiFattureCollegate", si debba escludere "DatiOrdineAcquisto"
        # e "DatiDDT". 
        # La cosa particolare e' che succede anche con la validazione di Zefiro, evidentemente
        # la validazione si collega ad un servizio on line di Agenzia delle Entrate che ha cambiato i parametri.
        # 
        if not 'DatiFattureCollegate' in self.json:
            # body: Dati riferimenti ordini
            for (i,v) in enumerate(self.json['righe']):
                if 'n_ord' in v or 'cup' in v or 'cig' in v or 'ordpa' in v:
                    t = self.addElement('DatiOrdineAcquisto', gen)
                    i0 = self.addElement('RiferimentoNumeroLinea', t, i+1, '%d')

                    # ora per default inserisco l'ordine inserito manualmente invece del riferimento interno
                    if 'ordpa' in v:
                        i0 = self.addElement('IdDocumento', t, v['ordpa'], '%s')
                        if 'dorpa' in v:
                            i0 = self.addElement('Data', t, v['dorpa'])
                    elif 'n_ord' in v:
                        i0 = self.addElement('IdDocumento', t, v['n_ord'], '%d')
                        i0 = self.addElement('Data', t, v['d_ord'])

                    if 'cms' in v:
                        i0 = self.addElement('CodiceCommessaConvenzione', t, v['cms'])
                    if 'cup' in v:
                        i0 = self.addElement('CodiceCUP', t, v['cup'])
                    if 'cig' in v:
                        i0 = self.addElement('CodiceCIG', t, v['cig'])

                    # se il cliente lo richiede si possono inserire altri campi
                    if 'numitem' in v:
                        i0 = self.addElement('NumItem', t, v['numitem'])

            # body: Dati ddt
            for (i,v) in enumerate(self.json['righe']):
                if 'n_ddt' in v:
                    t = self.addElement('DatiDDT', gen)
                    i0 = self.addElement('NumeroDDT', t, v['n_ddt'], '%d')
                    i0 = self.addElement('DataDDT', t, v['d_ddt'])
                    i0 = self.addElement('RiferimentoNumeroLinea', t, i+1, '%d')

        # body: Dati dettaglio linee 2.2
        beni = self.addElement('DatiBeniServizi', self.fat)
        numrow = 0
        aliva = float(self.ini.get('FatturaPA', 'AlIVA'))
        dec_prunit = int(self.ini.get('FatturaPA', 'DecPrunit'))
        for v in self.json['righe']:
            t = self.addElement('DettaglioLinee', beni) # 2.2.1
            numrow += 1
            i0 = self.addElement('NumeroLinea', t, numrow, '%d')
            if v['codart']:
                t1 = self.addElement('CodiceArticolo', t)
                i0 = self.addElement('CodiceTipo', t1, "Codice")
                i0 = self.addElement('CodiceValore', t1, v['codart'])
            i0 = self.addElement('Descrizione', t, v['descriz'])
            ## prima avevo disabilitato la quantita' e abilitato l'unita di misura
            ## adesso ho invertito dato che ci sono segnalazioni di errore sulle
            ## righe senza quantita'
            i0 = self.addElement('Quantita', t, v['qta'], '%.3f')
            if v['umis']:
                i0 = self.addElement('UnitaMisura', t, v['umis'])
            prezzo = v['valore']/v['qta'] if v['qta'] else 0
            i0 = self.addElement('PrezzoUnitario', t, calc_prezzo(prezzo))
            # i0 = self.addElement('PrezzoUnitario', t, prezzo, '%.7f') ## problemi di arrotondamento
            # i0 = self.addElement('PrezzoUnitario', t, round_down(prezzo, dec_prunit)) ## problemi di arrotondamento
            i0 = self.addElement('PrezzoTotale', t, v['valore'], '%.2f')
            if aliva and not v['valore'] and not v['iva']:
                v['iva'] = aliva
            i0 = self.addElement('AliquotaIVA', t, v['iva'], '%.2f')
            if 'ritenuta' in v:
                i0 = self.addElement('Ritenuta', t, 'SI')
            if not v['iva'] and (v['valore'] or not aliva):
                i0 = self.addElement('Natura', t, v['natura']) # 2.2.1.14
                if v['natura'] == 'N3.5' and self.json.get('cli_intento'):
                    i0 = self.addElement('AltriDatiGestionali', t) # 2.2.1.14
                    i1 = self.addElement('TipoDato', i0,'INTENTO') # 2.2.1.16.1
                    i1 = self.addElement('RiferimentoTesto', i0,self.json['cli_intento'][0]) # 2.2.1.16.2
                    i1 = self.addElement('RiferimentoData', i0,self.json['cli_intento'][1]) # 2.2.1.16.4
            if 'rifamm' in v:
                i0 = self.addElement('RiferimentoAmministrazione', t, v['rifamm'])
            if 'altridati' in v:
                    i0 = self.addElement('AltriDatiGestionali', t) # 2.2.1.14
                    i1 = self.addElement('TipoDato', i0, v['altridati'][0]) # 2.2.1.16.1
                    i1 = self.addElement('RiferimentoTesto', i0, v['altridati'][1]) # 2.2.1.16.2

        # body descrizioni fisse in fattura
        ad = self.ini.get('FatturaPA', 'AddDescriz')
        if ad and not self.autofat:
            ad = ad.split(';')
            for v in ad:
                t = self.addElement('DettaglioLinee', beni)
                numrow += 1
                i0 = self.addElement('NumeroLinea', t, numrow, '%d')
                i0 = self.addElement('Descrizione', t, v)

                i0 = self.addElement('Quantita', t, 0, '%.3f')
                i0 = self.addElement('PrezzoUnitario', t, 0, '%.5f')
                i0 = self.addElement('PrezzoTotale', t, 0, '%.2f')
                i0 = self.addElement('AliquotaIVA', t, 0, '%.2f')
                i0 = self.addElement('Natura', t, self.fuori_campo)
            n2_found = False
            for v in self.json['iva']:
                if not v['aliquota'] and v['natura'] == self.fuori_campo:
                    n2_found = True
            if not n2_found:
                self.json['iva'].append({
                    "aliquota" : 0 ,
                    "impo" : 0 ,
                    "iva" : 0 ,
                    "natura" : self.fuori_campo,
                    "norma" : ""
                })

        # body: riepilogo iva 2.2.2
        for (i,v) in enumerate(self.json['iva']):
            t = self.addElement('DatiRiepilogo', beni)
            i0 = self.addElement('AliquotaIVA', t, v['aliquota'], '%.2f')
            if not v['aliquota']:
                i0 = self.addElement('Natura', t, v['natura'])
            i0 = self.addElement('ImponibileImporto', t, v['impo'], '%.2f')
            i0 = self.addElement('Imposta', t, math_round(v['iva']), '%.2f')
            iva = 'D' if self.json['RegFis'] == 'RF16' else 'I'
            if v['aliquota'] != 0:
                iva = iva if self.json['NoSplitPay'] else 'S'
            i0 = self.addElement('EsigibilitaIVA', t, iva)
            if not v['aliquota']:
                if v['norma']:
                    i0 = self.addElement('RiferimentoNormativo', t, v['norma'])

        # body: Dati pagamento 2.4
        if self.json['scade']:
            pag = self.addElement('DatiPagamento', self.fat)
            t = self.addElement('CondizioniPagamento', pag, self.json['pagam'])
            for (i,v) in enumerate(self.json['scade']):
                t = self.addElement('DettaglioPagamento', pag)
                i0 = self.addElement('ModalitaPagamento', t, v['modo'])
                i0 = self.addElement('DataRiferimentoTerminiPagamento', t, self.json['d_doc'])
                i0 = self.addElement('GiorniTerminiPagamento', t, v['giorni'], '%d')
                i0 = self.addElement('DataScadenzaPagamento', t, v['data'])
                i0 = self.addElement('ImportoPagamento', t, v['impo'], '%.2f')
                if 'IstitutoFinanziario' in v:
                    i0 = self.addElement('IstitutoFinanziario', t, v['IstitutoFinanziario'][0:81])
                if 'iban' in v:
                    i0 = self.addElement('IBAN', t, v['iban'])
                if 'IstitutoFinanziario' in v:
                    i0 = self.addElement('ABI', t, v['ABI'])
                    i0 = self.addElement('CAB', t, v['CAB'])

        self.allega()

    def trasmissione(self, versione):
        # header: dati trasmissione
        t = self.addElement('DatiTrasmissione', self.hdr)
        i = self.addElement('IdTrasmittente', t)
        x = self.addElement('IdPaese', i, "IT")
        x = self.addElement('IdCodice', i, self.json['CodFis'])
        i = self.addElement('ProgressivoInvio', t, self.json['ProgInvio'])
        i = self.addElement('FormatoTrasmissione', t, versione)
        # i forfettari li posso escludere scrivendo 'NO' nel codice di interscambio
        if self.escluso:
            self.json['cli_dest'] = '0000000'
        i = self.addElement('CodiceDestinatario', t, self.json['cli_dest'])
        return t

    def allega(self):
        # body: allegati
        allegati = self.json.get('allegati', [])
        for ff in allegati:
            with open(ff['file'], 'rb') as f:
                bindata = f.read()
                f.close()
            b64 = base64.b64encode(bindata)
            if sys.version_info[0] > 2:
                b64 = str(b64, 'utf-8')
       
            allega = self.addElement('Allegati', self.fat)
            t = self.addElement('NomeAttachment', allega, ff['nome'])
            t = self.addElement('FormatoAttachment', allega, ff['formato'])
            if ff['descriz']:
                t = self.addElement('DescrizioneAttachment', allega, ff['descriz'])
            t = self.addElement('Attachment', allega, b64)

    def semplificata(self):

        # namespace fattura
        el = self.doc.createElementNS('http://ivaservizi.agenziaentrate.gov.it/docs/xsd/fatture/v1.0', 'ns2:FatturaElettronicaSemplificata')
        el.setAttribute("versione", "FSM10")
        el.setAttribute("xmlns:ns2", "http://ivaservizi.agenziaentrate.gov.it/docs/xsd/fatture/v1.0")
        self.doc.appendChild(el)
        self.hdr = el.appendChild(self.doc.createElement("FatturaElettronicaHeader"))
        self.fat = el.appendChild(self.doc.createElement("FatturaElettronicaBody"))

        t  = self.trasmissione('FSM10')

        t = self.addElement('CedentePrestatore', self.hdr)
        i1 = self.addElement('IdFiscaleIVA', t)
        i2 = self.addElement('IdPaese', i1, 'IT')
        i2 = self.addElement('IdCodice', i1, self.json['PIva'])
        i0 = self.addElement('CodiceFiscale', t, self.json['CodFis'])
        if self.json['TipoPers'] == 'S' or self.json['RagSoc'].strip():
            i0 = self.addElement('Denominazione', t, self.json['RagSoc'])
        else:
            i0 = self.addElement('Nome', t, self.json['Nome'])
            i0 = self.addElement('Cognome', t, self.json['Cogn'])
        i0 = self.addElement('Sede', t)
        i1 = self.addElement('Indirizzo', i0, self.json['via'])
        i1 = self.addElement('CAP', i0, self.json['cap'])
        i1 = self.addElement('Comune', i0, self.json['citta'])
        i1 = self.addElement('Provincia', i0, self.json['prov'])
        i1 = self.addElement('Nazione', i0, 'IT')
        if self.json['REA']: 
            i0 = self.addElement('IscrizioneREA', t)
            i1 = self.addElement('Ufficio', i0, self.json['REA'].split('-')[0])
            i1 = self.addElement('NumeroREA', i0, self.json['REA'].split('-')[1])
            if self.json['Capitale']: 
                i1 = self.addElement('CapitaleSociale', i0, self.json['Capitale'], '%.2f')
            i1 = self.addElement('SocioUnico', i0, 'SU' if self.json['SocioUnico'] else 'SM')
            i1 = self.addElement('StatoLiquidazione', i0, 'LN')
        i0 = self.addElement('RegimeFiscale', t, self.json['RegFis'])

        # header: cessionario committente
        t = self.addElement('CessionarioCommittente', self.hdr)
        i0 = self.addElement('IdentificativiFiscali', t)
        if self.json['cli_piva']:
            i1 = self.addElement('IdFiscaleIVA', i0)
            i2 = self.addElement('IdPaese', i1, self.json['cli_ivacee'])
            i2 = self.addElement('IdCodice', i1, self.json['cli_piva'])
            #  se c'e' un codice ficale valido diverso dalla partita iva lo includo sempre
            if self.json['cli_codfis'] and self.json['cli_codfis'].strip() != self.json['cli_piva']:
                i1 = self.addElement('CodiceFiscale', i0, self.json['cli_codfis'])
        else:
            i1 = self.addElement('CodiceFiscale', i0, self.json['cli_codfis'])

        # body: dati generali
        gen = self.addElement('DatiGenerali', self.fat)
        i0 = self.addElement('DatiGeneraliDocumento', gen)

        dd = {'TD01':'TD07', 'TD04':'TD08', 'TD05':'TD09', 'TD24':'TD07'} # corrispondenze tipi di documento
        i1 = self.addElement('TipoDocumento', i0, dd[self.json['t_doc']])
        i1 = self.addElement('Divisa', i0, 'EUR')
        i1 = self.addElement('Data', i0, self.json['d_doc'])
        if self.json['r_doc'] == 'A':
            i1 = self.addElement('Numero', i0, self.json['n_doc'], '%d')
        else:
            i1 = self.addElement('Numero', i0, "%d/%s" % (self.json['n_doc'],self.json['r_doc']))

        for v in self.json['righe']:
            t = self.addElement('DatiBeniServizi', self.fat)
            i0 = self.addElement('Descrizione', t, v['descriz'])
            iva = v['valore'] * v['iva'] / 100 
            i0 = self.addElement('Importo', t, v['valore'] + iva, '%.2f')
            i0 = self.addElement('DatiIVA', t)
            i1 = self.addElement('Aliquota', i0, v['iva'], '%.2f')
            if v['iva']:
                i0 = self.addElement('Natura', t, v['natura'])

        self.allega()

    def addElement(self, tag, node=None, text=None, format=None):
        if not node:
            node = self.doc.documentElement
        element = self.doc.createElement(tag)
        node.appendChild(element)
        if text != None:
            if format:
                text = format % text
            t = self.doc.createTextNode(text)
            element.appendChild(t)
        return element

    def write(self):
        return self.doc.toprettyxml(encoding="UTF-8")

    def save(self, path = None, path_tmp = None):
        
        path = self.json.get('path_out', path)
        path_tmp = self.json.get('path_tmp', path_tmp)

        # del 2019 in poi per il nome del file uso un codice codificato per gestire numeri grandi e gli invii multipli
        if int(self.json['d_doc'][0:4]) < 2019:
            try:
                noreg = self.ini.getboolean('VARIE', 'PYFPAREG')
            except:
                noreg = False
            if self.json['n_doc'] > 9999:
                noreg = True
            if noreg:
                file_name = 'IT%s_%05d.xml' % (self.json['CodFis'], self.json['n_doc'])
            else:
                file_name = 'IT%s_%s%04d.xml' % (self.json['CodFis'], self.json['r_doc'], self.json['n_doc'])
        else:
            prog = numprog(int(self.json['ID']), "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ")
            prog = ('00000'+prog)[-5:]
            file_name = 'IT%s_%s.xml' % (self.json['CodFis'], prog)

        fname = file_name if not path else os.path.join(path, file_name)
        s = self.write()
        f = open(fname,'wb+')
        f.write(s)
        f.close()
        
        # se ho passato la cartella dei files temporanei copio il file
        if path_tmp:
            shutil.copyfile(fname, os.path.join(path_tmp, file_name))

        print(file_name)

def open_config(fini):
    """Open an ini file with the name passed or same name of the python script. First search in the
    current dir and then in the dir of script"""
    if not fini:
        fini = os.path.basename(sys.argv[0]).rsplit('.', 1)[0] + '.ini'  # name of ini file = same of py script
    fname = os.path.join(os.path.curdir, fini)  # search first the curdir and then the script dir
    if not os.path.isfile( fname ):
        fname = sys.argv[0].rsplit('.', 1)[0] + '.ini'  # name of ini file = same of py script in current directory
    ini = cp.ConfigParser()
    ini.read(fname)
    return ini

def math_round(x):
    """Only in the particular case that the 3th decimal is 5 I add some value to ensure the exceded round"""
    if ('%.3f' % x)[-1] == '5':
        x += 0.0000001
    return x

def calc_prezzo(n):
    """fa una stringa con 7 decimali e toglie gli zeri finali fino a lasciare 2 decimali"""
    s = f"{n:.7f}"
    pos = s.index('.')
    while len(s) > pos+3 and s[-1] == '0':
        s = s[:-1]
    return s

def numprog(dec, digits = "0123456789ABCDEF"):
    # https://stackoverflow.com/questions/5796238/python-convert-decimal-to-hex
    # digits = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    # numbers = [0, 11, 16, 32, 33, 41, 45, 678, 574893]
    # print [numprog(x) for x in numbers]
    ll = len(digits)
    x = (dec % ll)
    rest = dec / ll
    if (rest == 0):
        return digits[int(x)]
    return numprog(rest, digits) + digits[int(x)]


if __name__ == '__main__':
    file_in = 'FatturaPa_json.txt'
    # file_in = 'FatturaPaAttach_json.txt'
    # file_in = 'FatturaPa_Sempl_json.txt'
    # file_in = 'FatturaPasall_json.txt'
    # file_in = 'FatturaPaEuro_json.txt'
    path_out = None
    path_tmp = None

    if len(sys.argv) > 1:
        file_in = sys.argv[1]

    if len(sys.argv) > 2:
        path_out = sys.argv[2]

    if len(sys.argv) > 3:
        path_tmp = sys.argv[3]

    ini = open_config('AGC.ini')

    f = FatPA(file_in, ini)
    f.save(path_out, path_tmp)
