#!/usr/bin/env python3
#
# author: Claudio Driussi
#
from xml.etree import ElementTree as ET
import sys
import os
import configparser
import json


class DatiFattura:

    def __init__(self, file_in, ini):
        # open data file and set configurationdata
        with open(file_in) as f:
            data = f.read()
            f.close()
        self.data_in = json.loads(data, encoding='CP1252')
        self.ini = ini
        self.tipo = 'ANN'

        # start xml
        ns = "http://ivaservizi.agenziaentrate.gov.it/docs/xsd/fatture/v2.0"
        ET.register_namespace("ns", ns)
        doc = ET.Element("{%s}DatiFattura" % ns)
        doc.attrib['versione'] = "DAT20"
        self.doc = doc

        root = doc
        self.header(root)

        # if self.data_in.has_key("DTE"):
        if "DTE" in self.data_in:
            self.tipo = 'DTE'
            self.anag_node = "CedentePrestatoreDTE"
            self.clifor_node = "CessionarioCommittenteDTE"
            self.fat_node = "DatiFatturaBodyDTE"
            self.DTE(root)
        # elif self.data_in.has_key("DTR"):
        elif "DTR" in self.data_in:
            self.tipo = 'DTR'
            self.anag_node = "CessionarioCommittenteDTR"
            self.clifor_node = "CedentePrestatoreDTR"
            self.fat_node = "DatiFatturaBodyDTR"
            self.DTR(root)
        else:
            self.ANN(root)

    def header(self, start_node):
        node = ET.SubElement(start_node, "DatiFatturaHeader")
        # optional values may not have the key
        if "ProgressivoInvio" in self.data_in:
            ET.SubElement(node, "ProgressivoInvio").text = self.data_in["ProgressivoInvio"]

    def DTE(self, start_node):
        node = ET.SubElement(start_node, "DTE")
        self.ditta(node)
        self.clifor(node, self.clifor_node)

    def DTR(self, start_node):
        node = ET.SubElement(start_node, "DTR")
        self.ditta(node)
        self.clifor(node, self.clifor_node)

    def ANN(self, start_node):
        node = ET.SubElement(start_node, "ANN")
        ET.SubElement(node, "IdFile").text = self.data_in["ANN"]

    def ditta(self, start_node):
        node = ET.SubElement(start_node, self.anag_node)
        self.anagrafica(node, self.data_in)

    def anagrafica(self, node, data, storg=None, rapfis=None):
        def sede(node, dd):
            if dd["Indirizzo"].strip():
                ET.SubElement(node, "Indirizzo").text = dd["Indirizzo"]
            ET.SubElement(node, "CAP").text = dd["CAP"]
            ET.SubElement(node, "Comune").text = dd["Comune"]
            if dd["Nazione"] == "IT":
                ET.SubElement(node, "Provincia").text = dd["Provincia"]
            ET.SubElement(node, "Nazione").text = dd["Nazione"]

        if "StabileOrganizzazione" in data:
            storg = data["StabileOrganizzazione"]
        if "RappresentanteFiscale" in data:
            rapfis = data["RappresentanteFiscale"]
        n1 = ET.SubElement(node, "IdentificativiFiscali")
        if "IdCodice" in data and data["IdCodice"]:
            n2 = ET.SubElement(n1,"IdFiscaleIVA")
            ET.SubElement(n2,"IdPaese").text = data["IdPaese"]
            ET.SubElement(n2,"IdCodice").text = data["IdCodice"]
        else:
            ET.SubElement(n1, "CodiceFiscale").text = data["CodiceFiscale"]
        n1 = ET.SubElement(node, "AltriDatiIdentificativi")
        if "Nome" in data and "Cognome" in data:
            ET.SubElement(n1, "Nome").text = data["Nome"]
            ET.SubElement(n1, "Cognome").text = data["Cognome"]
        else:
            ET.SubElement(n1, "Denominazione").text = data["Denominazione"]
        if data["CAP"] and data["Comune"]:
            n2 = ET.SubElement(n1, "Sede")
            sede(n2,data)
        if storg:
            n2 = ET.SubElement(n1, "StabileOrganizzazione")
            sede(n2,storg)
        if rapfis:
            n2 = ET.SubElement(n1, "RappresentanteFiscale")
            n3 = ET.SubElement(n2,"IdFiscaleIVA")
            ET.SubElement(n3,"IdPaese").text = rapfis["IdPaese"]
            ET.SubElement(n3,"IdCodice").text = rapfis["IdCodice"]
            if "Nome" in rapfis and "Cognome" in rapfis:
                ET.SubElement(n2, "Nome").text = rapfis["Nome"]
                ET.SubElement(n2, "Cognome").text = rapfis["Cognome"]
            else:
                ET.SubElement(n2, "Denominazione").text = rapfis["Denominazione"]

    def clifor(self, start_node, clifor_node):
        if len(self.data_in[self.tipo]) > 1000:
            raise ValueError("More than 1000 elements in node %s" % clifor_node)

        for cf in self.data_in[self.tipo]:
            node = ET.SubElement(start_node, clifor_node)
            self.anagrafica(node, cf)
            for doc in cf["DatiFatturaBody"]:
                node_doc = ET.SubElement(node, self.fat_node)
                self.documento(node_doc, doc)

    def documento(self, start_node, data):
        n1 = ET.SubElement(start_node, "DatiGenerali")
        ET.SubElement(n1, "TipoDocumento").text = data["TipoDocumento"]
        ET.SubElement(n1, "Data").text = data["Data"]
        ET.SubElement(n1, "Numero").text = data["Numero"]
        if self.tipo == 'DTR':
            ET.SubElement(n1, "DataRegistrazione").text = data["DataRegistrazione"]
        for civ in data["DatiRiepilogo"]:
            n1 = ET.SubElement(start_node, "DatiRiepilogo")
            ET.SubElement(n1, "ImponibileImporto").text = '%.2f' % civ["ImponibileImporto"]
            n2 = ET.SubElement(n1, "DatiIVA")
            ET.SubElement(n2, "Imposta").text = '%.2f' % civ["Imposta"]
            ET.SubElement(n2, "Aliquota").text = '%.2f' % civ["Aliquota"]
            if "Natura" in civ:
                ET.SubElement(n1, "Natura").text = civ["Natura"]
            if "Detraibile" in civ:
                ET.SubElement(n1, "Detraibile").text = '%.2f' % civ["Detraibile"]
            if "Deducibile" in civ:
                ET.SubElement(n1, "Deducibile").text = civ["Deducibile"]
            if "EsigibilitaIVA" in civ:
                ET.SubElement(n1, "EsigibilitaIVA").text = civ["EsigibilitaIVA"]

    def save(self, path_out = './'):
        indent(self.doc)
        print(self.data_in["NomeFile"])
        s = ET.tostring(self.doc, encoding='unicode')
        with open(path_out+self.data_in["NomeFile"], "w") as f:
            f.write(s)
            f.close()

def open_config(fini):
    """Open an ini file with the name passed or same name of the python script. First search in the
    current dir and then in the dir of script"""
    if not fini:
        fini = os.path.basename(sys.argv[0]).rsplit('.', 1)[0] + '.ini'  # name of ini file = same of py script
    fname = os.path.join(os.path.curdir, fini)  # search first the curdir and then the script dir
    if not os.path.isfile( fname ):
        fname = sys.argv[0].rsplit('.', 1)[0]
    ini = configparser.ConfigParser()
    ini.read(fname)
    return ini

def get_text(value):
    return value.__repr__()


def indent(elem, level=0):
    """Pretty print"""
    i = "\n" + level*"  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            indent(elem, level+1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


if __name__ == '__main__':
    file_in = 'IT01463820934_DF_19001.json'
    path_out = './'

    if len(sys.argv) > 1:
        file_in = sys.argv[1]

    if len(sys.argv) > 2:
        path_out = sys.argv[2]

    ini = open_config('AGC.ini')

    f = DatiFattura(file_in, ini)
    f.save(path_out)
