#!/usr/bin/env python3
#
# author: Claudio Driussi
#
import sys
import os
import json
from xml.etree import ElementTree as ET
import xmlschema

class Validate:

    def __init__(self, xsd_file):
        self.schema = xmlschema.XMLSchema(xsd_file)
        self.errors = []

    def validate(self, xml_files):
        """Fa la validazione di un file o una directory, se passo un nome che 
        finisce con .xml e' un file, altrimenti una directory"""
        
        # trova i files
        extension = os.path.splitext(xml_files)[1].upper()
        if extension == '.XML':
            files = [xml_files]
        else:
            files = []
            dirfiles = os.listdir(xml_files)
            for file in dirfiles:
                filename = os.path.join(xml_files, file)
                if os.path.isfile(filename):
                    if os.path.splitext(filename)[1].upper() == '.XML':
                        files.append(filename) 

        # esegue la validazione
        for file in files:
            self._validate(file)

    def _validate(self, xml_file):
        """valida un singolo file, se trova errori aggiunge ragione sociale ed 
        estremi documento"""
        err = {}
        err['file'] = xml_file
        err['errors'] = []
        validation_error_iterator = self.schema.iter_errors(xml_file)
        for validation_error in validation_error_iterator:
            print(f"{xml_file}: {validation_error.message}")
            if not err['errors']:
                self.read_fpa(xml_file, err)
            e = {}
            e['message'] = validation_error.message
            if validation_error.elem is not None:
                e['tag'] = validation_error.elem.tag
            if validation_error.path is not None:
                e['path'] = validation_error.path
            err['errors'].append(e)           
        # aggiunge i dati del file solo se ci sono errori
        if err['errors']:
            self.errors.append(err)

    def read_fpa(self, xml_file, err):
        tree = ET.parse(xml_file)
        root = tree.getroot()
        head = root.find('FatturaElettronicaHeader')
        body = root.find('FatturaElettronicaBody')
        err['fornitore'] = self.get_key(head, 'CedentePrestatore/DatiAnagrafici/Anagrafica/Denominazione//////////', '')
        if not err['fornitore']:
            n1 = self.get_key(head, 'CedentePrestatore/DatiAnagrafici/Anagrafica/Nome//////////', '')
            n2 = self.get_key(head, 'CedentePrestatore/DatiAnagrafici/Anagrafica/Cognome/////////', '')
            err['fornitore'] = f"{n1} {n2}"
        err['cliente'] = self.get_key(head, 'CessionarioCommittente/DatiAnagrafici/Anagrafica/Denominazione//////////', '')
        if not err['cliente']:
            n1 = self.get_key(head, 'CessionarioCommittente/DatiAnagrafici/Anagrafica/Nome//////////', '')
            n2 = self.get_key(head, 'CessionarioCommittente/DatiAnagrafici/Anagrafica/Cognome/////////', '')
            err['cliente'] = f"{n1} {n2}"
        err['t_doc'] = self.get_key(body, 'DatiGenerali/DatiGeneraliDocumento/TipoDocumento///////////', '')
        err['d_doc'] = self.get_key(body, 'DatiGenerali/DatiGeneraliDocumento/Data///////////', '')
        err['n_doc'] = self.get_key(body, 'DatiGenerali/DatiGeneraliDocumento/Numero///////////', '')

    def get_key(self, element, key, default=""):
        key = key.rstrip('/')
        val = element.find(key)
        if val is None:
            return default
        return val.text

    def save(self, err_file):
        with open(err_file, 'w') as f:
            json.dump(self.errors, f,indent=2)

if __name__ == '__main__':
    xml_files = 'validate'
    # xml_files = 'validate/ITDRSCLD56M29F750D_0000E.xml'
    xsd_file = os.path.join(os.path.dirname(sys.argv[0]), 'Schema_VFPR12.xsd') 
    err_file = 'fpa_errors.json'

    if len(sys.argv) > 1:
        xml_files = sys.argv[1]
    if len(sys.argv) > 2:
        err_file = sys.argv[2]
    if len(sys.argv) > 3:
        xsd_file = sys.argv[3]

    valid = Validate(xsd_file)
    valid.validate(xml_files)
    valid.save(err_file)
